@@ -133,7 +133,6 @@ def novaclient(context, privileged_user=False, timeout=None, api_version=None):
 
 class API(base.Base):
     """API for interacting with novaclient."""
-    NotFound = nova_exceptions.NotFound
 
     def __init__(self):
         self.message_api = message_api.API()
@@ -227,7 +226,7 @@ class API(base.Base):
         return result
 
     def reimage_volume(self, context, server_ids, volume_id):
-        api_version = '2.91'
+        api_version = '2.93'
         events = [self._get_volume_reimaged_event(server_id, volume_id)
                   for server_id in server_ids]
         result = self._send_events(context, events, api_version=api_version)
@@ -236,11 +235,5 @@ class API(base.Base):
                 context,
                 message_field.Action.REIMAGE_VOLUME,
                 resource_uuid=volume_id,
-                detail=message_field.Detail.NOTIFY_COMPUTE_SERVICE_FAILED)
+                detail=message_field.Detail.REIMAGE_VOLUME_FAILED)
         return result
-
-    @staticmethod
-    def get_server_volume(context, server_id, volume_id):
-        # Use microversion that includes attachment_id
-        nova = novaclient(context, api_version='2.89')
-        return nova.volumes.get_server_volume(server_id, volume_id)
