@@ -34,6 +34,7 @@ from sqlalchemy.engine import reflection
 
 import cinder.db.legacy_migrations
 from cinder.db import migration
+from cinder.db.sqlalchemy import api
 from cinder.db.sqlalchemy import models
 from cinder.tests import fixtures as cinder_fixtures
 from cinder.tests.unit import utils as test_utils
@@ -58,6 +59,7 @@ class CinderModelsMigrationsSync(test_migrations.ModelsMigrationsSync):
         self.useFixture(cinder_fixtures.StandardLogging())
 
         self.engine = enginefacade.writer.get_engine()
+        self.patch(api, 'get_engine', self.get_engine)
 
     def db_sync(self, engine):
         migration.db_sync(engine=self.engine)
@@ -135,10 +137,12 @@ class MigrationsWalk(
     # Migrations can take a long time, particularly on underpowered CI nodes.
     # Give them some breathing room.
     TIMEOUT_SCALING_FACTOR = 4
+    BOOL_TYPE = sqlalchemy.types.BOOLEAN
 
     def setUp(self):
         super().setUp()
         self.engine = enginefacade.writer.get_engine()
+        self.patch(api, 'get_engine', lambda: self.engine)
         self.config = migration._find_alembic_conf()
         self.init_version = migration.ALEMBIC_INIT_VERSION
 
@@ -198,6 +202,31 @@ class MigrationsWalk(
         ).get_current_head()
         self.assertEqual(head, migration.db_version())
 
+    def _pre_upgrade_c92a3e68beed(self, connection):
+        """Test shared_targets is nullable."""
+        table = db_utils.get_table(connection, 'volumes')
+        self._previous_type = type(table.c.shared_targets.type)
+
+    def _check_c92a3e68beed(self, connection):
+        """Test shared_targets is nullable."""
+        table = db_utils.get_table(connection, 'volumes')
+        self.assertIn('shared_targets', table.c)
+        # Type hasn't changed
+        self.assertIsInstance(table.c.shared_targets.type, self._previous_type)
+        # But it's nullable
+        self.assertTrue(table.c.shared_targets.nullable)
+
+    def _check_daa98075b90d(self, connection):
+        """Test resources have indexes."""
+        for table in ('groups', 'group_snapshots', 'volumes', 'snapshots',
+                      'backups'):
+            db_utils.index_exists(connection,
+                                  table,
+                                  f'{table}_deleted_project_id_idx')
+
+        db_utils.index_exists(connection,
+                              'volumes', 'volumes_deleted_host_idx')
+
 
 class TestMigrationsWalkSQLite(
     MigrationsWalk,
