@@ -14,10 +14,11 @@
 #    under the License.
 
 import json
-import mock
+from unittest import mock
+
 import requests
 
-from cinder import test
+from cinder.tests.unit import test
 from cinder.tests.unit.volume.drivers.fusionstorage import test_utils
 from cinder.volume.drivers.fusionstorage import fs_client
 
