@@ -1,4 +1,4 @@
-# Copyright (C) 2020, 2021, Hitachi, Ltd.
+# Copyright (C) 2020, 2023, Hitachi, Ltd.
 #
 #    Licensed under the Apache License, Version 2.0 (the "License"); you may
 #    not use this file except in compliance with the License. You may obtain
@@ -18,6 +18,7 @@ import functools
 from unittest import mock
 
 from oslo_config import cfg
+from oslo_utils import units
 import requests
 from requests import models
 
@@ -36,6 +37,7 @@ from cinder.volume import configuration as conf
 from cinder.volume import driver
 from cinder.volume.drivers.hitachi import hbsd_common
 from cinder.volume.drivers.hitachi import hbsd_fc
+from cinder.volume.drivers.hitachi import hbsd_replication
 from cinder.volume.drivers.hitachi import hbsd_rest
 from cinder.volume.drivers.hitachi import hbsd_rest_api
 from cinder.volume.drivers.hitachi import hbsd_rest_fc
@@ -182,6 +184,16 @@ GET_HOST_WWNS_RESULT = {
     ],
 }
 
+GET_HOST_GROUPS_RESULT_TEST = {
+    "data": [
+        {
+            "hostGroupNumber": 0,
+            "portId": CONFIG_MAP['port_id'],
+            "hostGroupName": CONFIG_MAP['host_grp_name'],
+        },
+    ],
+}
+
 COMPLETED_SUCCEEDED_RESULT = {
     "status": "Completed",
     "state": "Succeeded",
@@ -204,6 +216,9 @@ GET_LDEV_RESULT = {
     "blockCapacity": 2097152,
     "attributes": ["CVS", "HDP"],
     "status": "NML",
+    "poolId": 30,
+    "dataReductionStatus": "DISABLED",
+    "dataReductionMode": "disabled",
 }
 
 GET_LDEV_RESULT_MAPPED = {
@@ -228,6 +243,16 @@ GET_LDEV_RESULT_PAIR = {
     "status": "NML",
 }
 
+GET_LDEV_RESULT_PAIR_STATUS_TEST = {
+    "emulationType": "OPEN-V-CVS",
+    "blockCapacity": 2097152,
+    "attributes": ["CVS", "HDP", "HTI"],
+    "status": "TEST",
+    "poolId": 30,
+    "dataReductionStatus": "REHYDRATING",
+    "dataReductionMode": "disabled"
+}
+
 GET_POOL_RESULT = {
     "availableVolumeCapacity": 480144,
     "totalPoolCapacity": 507780,
@@ -275,6 +300,10 @@ GET_POOLS_RESULT = {
         {
             "poolId": 30,
             "poolName": CONFIG_MAP['pool_name'],
+            "availableVolumeCapacity": 480144,
+            "totalPoolCapacity": 507780,
+            "totalLocatedCapacity": 71453172,
+            "virtualVolumeCapacityRate": -1,
         },
     ],
 }
@@ -302,6 +331,16 @@ GET_HOST_GROUPS_RESULT = {
     ],
 }
 
+GET_HOST_GROUPS_RESULT_PAIR = {
+    "data": [
+        {
+            "hostGroupNumber": 1,
+            "portId": CONFIG_MAP['port_id'],
+            "hostGroupName": "HBSD-pair00",
+        },
+    ],
+}
+
 GET_LDEVS_RESULT = {
     "data": [
         {
@@ -453,7 +492,7 @@ class HBSDRESTFCDriverTest(test.TestCase):
         self.configuration.driver_ssl_cert_verify = False
 
         self.configuration.hitachi_storage_id = CONFIG_MAP['serial']
-        self.configuration.hitachi_pool = "30"
+        self.configuration.hitachi_pools = ["30"]
         self.configuration.hitachi_snap_pool = None
         self.configuration.hitachi_ldev_range = "0-1"
         self.configuration.hitachi_target_ports = [CONFIG_MAP['port_id']]
@@ -465,6 +504,7 @@ class HBSDRESTFCDriverTest(test.TestCase):
         self.configuration.hitachi_copy_check_interval = 3
         self.configuration.hitachi_async_copy_check_interval = 10
         self.configuration.hitachi_port_scheduler = False
+        self.configuration.hitachi_group_name_format = None
 
         self.configuration.san_login = CONFIG_MAP['user_id']
         self.configuration.san_password = CONFIG_MAP['user_pass']
@@ -475,7 +515,6 @@ class HBSDRESTFCDriverTest(test.TestCase):
         self.configuration.hitachi_rest_disable_io_wait = True
         self.configuration.hitachi_rest_tcp_keepalive = True
         self.configuration.hitachi_discard_zero_page = True
-        self.configuration.hitachi_rest_number = "0"
         self.configuration.hitachi_lun_timeout = hbsd_rest._LUN_TIMEOUT
         self.configuration.hitachi_lun_retry_interval = (
             hbsd_rest._LUN_RETRY_INTERVAL)
@@ -523,6 +562,21 @@ class HBSDRESTFCDriverTest(test.TestCase):
         self.configuration.chap_username = CONFIG_MAP['auth_user']
         self.configuration.chap_password = CONFIG_MAP['auth_password']
 
+        self.configuration.hitachi_replication_number = 0
+        self.configuration.hitachi_pair_target_number = 0
+        self.configuration.hitachi_rest_pair_target_ports = []
+        self.configuration.hitachi_quorum_disk_id = ''
+        self.configuration.hitachi_mirror_copy_speed = ''
+        self.configuration.hitachi_mirror_storage_id = ''
+        self.configuration.hitachi_mirror_pool = ''
+        self.configuration.hitachi_mirror_ldev_range = ''
+        self.configuration.hitachi_mirror_target_ports = ''
+        self.configuration.hitachi_mirror_rest_user = ''
+        self.configuration.hitachi_mirror_rest_password = ''
+        self.configuration.hitachi_mirror_rest_api_ip = ''
+        self.configuration.hitachi_set_mirror_reserve_attribute = ''
+        self.configuration.hitachi_path_group_id = ''
+
         self.configuration.safe_get = self._fake_safe_get
 
         CONF = cfg.CONF
@@ -547,7 +601,8 @@ class HBSDRESTFCDriverTest(test.TestCase):
             configuration=self.configuration)
         request.side_effect = [FakeResponse(200, POST_SESSIONS_RESULT),
                                FakeResponse(200, GET_PORTS_RESULT),
-                               FakeResponse(200, GET_HOST_WWNS_RESULT)]
+                               FakeResponse(200, GET_HOST_WWNS_RESULT),
+                               FakeResponse(200, GET_HOST_GROUPS_RESULT_PAIR)]
         self.driver.do_setup(None)
         self.driver.check_for_setup_error()
         self.driver.local_path(None)
@@ -574,13 +629,14 @@ class HBSDRESTFCDriverTest(test.TestCase):
         self._setup_config()
         request.side_effect = [FakeResponse(200, POST_SESSIONS_RESULT),
                                FakeResponse(200, GET_PORTS_RESULT),
-                               FakeResponse(200, GET_HOST_WWNS_RESULT)]
+                               FakeResponse(200, GET_HOST_WWNS_RESULT),
+                               FakeResponse(200, GET_HOST_GROUPS_RESULT_PAIR)]
         drv.do_setup(None)
         self.assertEqual(
             {CONFIG_MAP['port_id']: CONFIG_MAP['target_wwn']},
             drv.common.storage_info['wwns'])
         self.assertEqual(1, brick_get_connector_properties.call_count)
-        self.assertEqual(3, request.call_count)
+        self.assertEqual(4, request.call_count)
         # stop the Loopingcall within the do_setup treatment
         self.driver.common.client.keep_session_loop.stop()
         self.driver.common.client.keep_session_loop.wait()
@@ -601,17 +657,67 @@ class HBSDRESTFCDriverTest(test.TestCase):
                                FakeResponse(200, NOTFOUND_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
-                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(200, GET_HOST_GROUPS_RESULT_PAIR)]
         drv.do_setup(None)
         self.assertEqual(
             {CONFIG_MAP['port_id']: CONFIG_MAP['target_wwn']},
             drv.common.storage_info['wwns'])
         self.assertEqual(1, brick_get_connector_properties.call_count)
-        self.assertEqual(8, request.call_count)
+        self.assertEqual(9, request.call_count)
         # stop the Loopingcall within the do_setup treatment
         self.driver.common.client.keep_session_loop.stop()
         self.driver.common.client.keep_session_loop.wait()
 
+    @mock.patch.object(requests.Session, "request")
+    @mock.patch.object(
+        volume_utils, 'brick_get_connector_properties',
+        side_effect=_brick_get_connector_properties)
+    def test_do_setup_create_hg_format(
+            self, brick_get_connector_properties, request):
+        drv = hbsd_fc.HBSDFCDriver(configuration=self.configuration)
+        self._setup_config()
+        self.configuration.hitachi_group_name_format = (
+            'HBSD-{wwn}-{host}-_:.@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@')
+        request.side_effect = [FakeResponse(200, POST_SESSIONS_RESULT),
+                               FakeResponse(200, GET_PORTS_RESULT),
+                               FakeResponse(200, NOTFOUND_RESULT),
+                               FakeResponse(200, NOTFOUND_RESULT),
+                               FakeResponse(200, NOTFOUND_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(200, GET_HOST_GROUPS_RESULT_PAIR)]
+        drv.do_setup(None)
+        self.assertEqual(
+            {CONFIG_MAP['port_id']: CONFIG_MAP['target_wwn']},
+            drv.common.storage_info['wwns'])
+        self.assertEqual(1, brick_get_connector_properties.call_count)
+        self.assertEqual(9, request.call_count)
+        # stop the Loopingcall within the do_setup treatment
+        self.driver.common.client.keep_session_loop.stop()
+        self.driver.common.client.keep_session_loop.wait()
+
+    @mock.patch.object(requests.Session, "request")
+    @mock.patch.object(
+        volume_utils, 'brick_get_connector_properties',
+        side_effect=_brick_get_connector_properties)
+    def test_do_setup_create_hg_format_error(
+            self, brick_get_connector_properties, request):
+        drv = hbsd_fc.HBSDFCDriver(configuration=self.configuration)
+        self._setup_config()
+        self.configuration.hitachi_group_name_format = '{host}-{wwn}'
+        request.side_effect = [FakeResponse(200, POST_SESSIONS_RESULT),
+                               FakeResponse(200, GET_PORTS_RESULT),
+                               FakeResponse(200, NOTFOUND_RESULT),
+                               FakeResponse(200, NOTFOUND_RESULT),
+                               FakeResponse(200, NOTFOUND_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
+
+        self.assertRaises(exception.VolumeDriverException, drv.do_setup, None)
+
     @mock.patch.object(requests.Session, "request")
     @mock.patch.object(
         volume_utils, 'brick_get_connector_properties',
@@ -633,13 +739,14 @@ class HBSDRESTFCDriverTest(test.TestCase):
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
-                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(200, GET_HOST_GROUPS_RESULT_PAIR)]
         drv.do_setup(None)
         self.assertEqual(
             {CONFIG_MAP['port_id']: CONFIG_MAP['target_wwn']},
             drv.common.storage_info['wwns'])
         self.assertEqual(1, brick_get_connector_properties.call_count)
-        self.assertEqual(9, request.call_count)
+        self.assertEqual(10, request.call_count)
         # stop the Loopingcall within the do_setup treatment
         self.driver.common.client.keep_session_loop.stop()
         self.driver.common.client.keep_session_loop.wait()
@@ -653,19 +760,20 @@ class HBSDRESTFCDriverTest(test.TestCase):
         drv = hbsd_fc.HBSDFCDriver(
             configuration=self.configuration)
         self._setup_config()
-        tmp_pool = self.configuration.hitachi_pool
-        self.configuration.hitachi_pool = CONFIG_MAP['pool_name']
+        tmp_pools = self.configuration.hitachi_pools
+        self.configuration.hitachi_pools = [CONFIG_MAP['pool_name']]
         request.side_effect = [FakeResponse(200, POST_SESSIONS_RESULT),
                                FakeResponse(200, GET_POOLS_RESULT),
                                FakeResponse(200, GET_PORTS_RESULT),
-                               FakeResponse(200, GET_HOST_WWNS_RESULT)]
+                               FakeResponse(200, GET_HOST_WWNS_RESULT),
+                               FakeResponse(200, GET_HOST_GROUPS_RESULT_PAIR)]
         drv.do_setup(None)
         self.assertEqual(
             {CONFIG_MAP['port_id']: CONFIG_MAP['target_wwn']},
             drv.common.storage_info['wwns'])
         self.assertEqual(1, brick_get_connector_properties.call_count)
-        self.assertEqual(4, request.call_count)
-        self.configuration.hitachi_pool = tmp_pool
+        self.assertEqual(5, request.call_count)
+        self.configuration.hitachi_pools = tmp_pools
         # stop the Loopingcall within the do_setup treatment
         self.driver.common.client.keep_session_loop.stop()
         self.driver.common.client.keep_session_loop.wait()
@@ -673,10 +781,32 @@ class HBSDRESTFCDriverTest(test.TestCase):
     @mock.patch.object(requests.Session, "request")
     def test_create_volume(self, request):
         request.return_value = FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
+        self.driver.common._stats = {}
+        self.driver.common._stats['pools'] = [
+            {'location_info': {'pool_id': 30}}]
         ret = self.driver.create_volume(fake_volume.fake_volume_obj(self.ctxt))
         self.assertEqual('1', ret['provider_location'])
         self.assertEqual(2, request.call_count)
 
+    @mock.patch.object(requests.Session, "request")
+    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
+    def test_create_volume_deduplication_compression(
+            self, get_volume_type_extra_specs, request):
+        extra_specs = {'hbsd:capacity_saving': 'deduplication_compression'}
+        get_volume_type_extra_specs.return_value = extra_specs
+        request.return_value = FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
+        self.driver.common._stats = {}
+        self.driver.common._stats['pools'] = [
+            {'location_info': {'pool_id': 30}}]
+        ret = self.driver.create_volume(TEST_VOLUME[3])
+        args, kwargs = request.call_args_list[0]
+        body = kwargs['json']
+        self.assertEqual(body.get('dataReductionMode'),
+                         'compression_deduplication')
+        self.assertEqual('1', ret['provider_location'])
+        self.assertEqual(1, get_volume_type_extra_specs.call_count)
+        self.assertEqual(2, request.call_count)
+
     @reduce_retrying_time
     @mock.patch.object(requests.Session, "request")
     def test_create_volume_timeout(self, request):
@@ -686,6 +816,9 @@ class HBSDRESTFCDriverTest(test.TestCase):
             500, ERROR_RESULT,
             headers={'Content-Type': 'json'})
 
+        self.driver.common._stats = {}
+        self.driver.common._stats['pools'] = [
+            {'location_info': {'pool_id': 30}}]
         self.assertRaises(exception.VolumeDriverException,
                           self.driver.create_volume,
                           fake_volume.fake_volume_obj(self.ctxt))
@@ -740,25 +873,99 @@ class HBSDRESTFCDriverTest(test.TestCase):
     @mock.patch.object(requests.Session, "request")
     def test_get_volume_stats(
             self, request, get_filter_function, get_goodness_function):
-        request.return_value = FakeResponse(200, GET_POOL_RESULT)
+        request.return_value = FakeResponse(200, GET_POOLS_RESULT)
         get_filter_function.return_value = None
         get_goodness_function.return_value = None
         stats = self.driver.get_volume_stats(True)
         self.assertEqual('Hitachi', stats['vendor_name'])
+        self.assertEqual(self.configuration.volume_backend_name,
+                         stats["pools"][0]['pool_name'])
+        self.assertEqual(self.configuration.reserved_percentage,
+                         stats["pools"][0]['reserved_percentage'])
+        self.assertTrue(stats["pools"][0]['thin_provisioning_support'])
+        self.assertFalse(stats["pools"][0]['thick_provisioning_support'])
         self.assertTrue(stats["pools"][0]['multiattach'])
+        self.assertTrue(stats["pools"][0]['consistencygroup_support'])
+        self.assertTrue(stats["pools"][0]['consistent_group_snapshot_enabled'])
+        self.assertEqual(self.configuration.max_over_subscription_ratio,
+                         stats["pools"][0]['max_over_subscription_ratio'])
+        self.assertEqual(
+            GET_POOL_RESULT['totalPoolCapacity'] // units.Ki,
+            stats["pools"][0]['total_capacity_gb'])
+        self.assertEqual(
+            GET_POOL_RESULT['availableVolumeCapacity'] // units.Ki,
+            stats["pools"][0]['free_capacity_gb'])
+        self.assertEqual(
+            GET_POOL_RESULT['totalLocatedCapacity'] // units.Ki,
+            stats["pools"][0]['provisioned_capacity_gb'])
+        self.assertEqual('up', stats["pools"][0]['backend_state'])
         self.assertEqual(1, request.call_count)
         self.assertEqual(1, get_filter_function.call_count)
         self.assertEqual(1, get_goodness_function.call_count)
 
+    @mock.patch.object(driver.FibreChannelDriver, "get_goodness_function")
+    @mock.patch.object(driver.FibreChannelDriver, "get_filter_function")
+    @mock.patch.object(hbsd_rest.HBSDREST, "get_pool_info")
+    def test_get_volume_stats_error(
+            self, get_pool_info, get_filter_function, get_goodness_function):
+        get_pool_info.side_effect = exception.VolumeDriverException(data='')
+        get_filter_function.return_value = None
+        get_goodness_function.return_value = None
+        stats = self.driver.get_volume_stats(True)
+        self.assertEqual('Hitachi', stats['vendor_name'])
+        self.assertEqual(self.configuration.volume_backend_name,
+                         stats["pools"][0]['pool_name'])
+        self.assertEqual(self.configuration.reserved_percentage,
+                         stats["pools"][0]['reserved_percentage'])
+        self.assertTrue(stats["pools"][0]['thin_provisioning_support'])
+        self.assertFalse(stats["pools"][0]['thick_provisioning_support'])
+        self.assertTrue(stats["pools"][0]['multiattach'])
+        self.assertTrue(stats["pools"][0]['consistencygroup_support'])
+        self.assertTrue(stats["pools"][0]['consistent_group_snapshot_enabled'])
+        self.assertEqual(self.configuration.max_over_subscription_ratio,
+                         stats["pools"][0]['max_over_subscription_ratio'])
+        self.assertEqual(0, stats["pools"][0]['total_capacity_gb'])
+        self.assertEqual(0, stats["pools"][0]['free_capacity_gb'])
+        self.assertEqual(0, stats["pools"][0]['provisioned_capacity_gb'])
+        self.assertEqual('down', stats["pools"][0]['backend_state'])
+        self.assertEqual(1, get_filter_function.call_count)
+        self.assertEqual(1, get_goodness_function.call_count)
+
     @mock.patch.object(requests.Session, "request")
+    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
     @mock.patch.object(sqlalchemy_api, 'volume_get', side_effect=_volume_get)
-    def test_create_snapshot(self, volume_get, request):
+    def test_create_snapshot(
+            self, volume_get, get_volume_type_extra_specs, request):
         request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(200, GET_SNAPSHOTS_RESULT)]
+        get_volume_type_extra_specs.return_value = {}
+        self.driver.common._stats = {}
+        self.driver.common._stats['pools'] = [
+            {'location_info': {'pool_id': 30}}]
         ret = self.driver.create_snapshot(TEST_SNAPSHOT[0])
         self.assertEqual('1', ret['provider_location'])
+        self.assertEqual(1, get_volume_type_extra_specs.call_count)
+        self.assertEqual(4, request.call_count)
+
+    @mock.patch.object(requests.Session, "request")
+    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
+    @mock.patch.object(sqlalchemy_api, 'volume_get', side_effect=_volume_get)
+    def test_create_snapshot_dedup_false(
+            self, volume_get, get_volume_type_extra_specs, request):
+        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(200, GET_SNAPSHOTS_RESULT)]
+        get_volume_type_extra_specs.return_value = {'hbsd:capacity_saving':
+                                                    'disable'}
+        self.driver.common._stats = {}
+        self.driver.common._stats['pools'] = [
+            {'location_info': {'pool_id': 30}}]
+        ret = self.driver.create_snapshot(TEST_SNAPSHOT[0])
+        self.assertEqual('1', ret['provider_location'])
+        self.assertEqual(1, get_volume_type_extra_specs.call_count)
         self.assertEqual(4, request.call_count)
 
     @mock.patch.object(requests.Session, "request")
@@ -772,9 +979,13 @@ class HBSDRESTFCDriverTest(test.TestCase):
                                FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
         self.driver.delete_snapshot(TEST_SNAPSHOT[0])
-        self.assertEqual(10, request.call_count)
+        self.assertEqual(14, request.call_count)
 
     @mock.patch.object(requests.Session, "request")
     def test_delete_snapshot_no_pair(self, request):
@@ -787,26 +998,40 @@ class HBSDRESTFCDriverTest(test.TestCase):
         self.assertEqual(4, request.call_count)
 
     @mock.patch.object(requests.Session, "request")
-    def test_create_cloned_volume(self, request):
+    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
+    def test_create_cloned_volume(
+            self, get_volume_type_extra_specs, request):
         request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(200, GET_SNAPSHOTS_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
+        get_volume_type_extra_specs.return_value = {}
+        self.driver.common._stats = {}
+        self.driver.common._stats['pools'] = [
+            {'location_info': {'pool_id': 30}}]
         vol = self.driver.create_cloned_volume(TEST_VOLUME[0], TEST_VOLUME[1])
         self.assertEqual('1', vol['provider_location'])
+        self.assertEqual(1, get_volume_type_extra_specs.call_count)
         self.assertEqual(5, request.call_count)
 
     @mock.patch.object(requests.Session, "request")
-    def test_create_volume_from_snapshot(self, request):
+    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
+    def test_create_volume_from_snapshot(
+            self, get_volume_type_extra_specs, request):
         request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(200, GET_SNAPSHOTS_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
+        self.driver.common._stats = {}
+        get_volume_type_extra_specs.return_value = {}
+        self.driver.common._stats['pools'] = [
+            {'location_info': {'pool_id': 30}}]
         vol = self.driver.create_volume_from_snapshot(
             TEST_VOLUME[0], TEST_SNAPSHOT[0])
         self.assertEqual('1', vol['provider_location'])
+        self.assertEqual(1, get_volume_type_extra_specs.call_count)
         self.assertEqual(5, request.call_count)
 
     @mock.patch.object(fczm_utils, "add_fc_zone")
@@ -877,6 +1102,32 @@ class HBSDRESTFCDriverTest(test.TestCase):
         self.assertEqual(5, request.call_count)
         self.assertEqual(1, add_fc_zone.call_count)
 
+    @mock.patch.object(fczm_utils, "add_fc_zone")
+    @mock.patch.object(requests.Session, "request")
+    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
+    def test_create_target_to_storage_return(
+            self, get_volume_type_extra_specs, request, add_fc_zone):
+        self.configuration.hitachi_zoning_request = True
+        self.driver.common._lookup_service = FakeLookupService()
+        extra_specs = {"hbsd:target_ports": "CL1-A"}
+        get_volume_type_extra_specs.return_value = extra_specs
+        request.side_effect = [
+            FakeResponse(200, NOTFOUND_RESULT),
+            FakeResponse(200, NOTFOUND_RESULT),
+            FakeResponse(200, GET_HOST_GROUPS_RESULT),
+            FakeResponse(200, NOTFOUND_RESULT),
+            FakeResponse(400, GET_HOST_GROUPS_RESULT),
+            FakeResponse(200, GET_HOST_GROUPS_RESULT_TEST),
+            FakeResponse(200, GET_HOST_GROUPS_RESULT_TEST),
+        ]
+        self.assertRaises(exception.VolumeDriverException,
+                          self.driver.initialize_connection,
+                          TEST_VOLUME[1],
+                          DEFAULT_CONNECTOR)
+        self.assertEqual(1, get_volume_type_extra_specs.call_count)
+        self.assertEqual(10, request.call_count)
+        self.assertEqual(0, add_fc_zone.call_count)
+
     @mock.patch.object(fczm_utils, "remove_fc_zone")
     @mock.patch.object(requests.Session, "request")
     def test_terminate_connection(self, request, remove_fc_zone):
@@ -1024,14 +1275,80 @@ class HBSDRESTFCDriverTest(test.TestCase):
             self.driver.unmanage_snapshot,
             TEST_SNAPSHOT[0])
 
-    def test_retype(self):
-        new_specs = {'hbsd:test': 'test'}
+    @mock.patch.object(requests.Session, "request")
+    def test_retype(self, request):
+        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
+        host = {
+            'capabilities': {
+                'location_info': {
+                    'pool_id': 30,
+                },
+            },
+        }
+        new_type = {'extra_specs': {
+            'hbsd:capacity_saving': 'deduplication_compression'}}
+        old_specs = {'hbsd:capacity_saving': 'disable'}
+        new_specs = {'hbsd:capacity_saving': 'deduplication_compression'}
+        old_type_ref = volume_types.create(self.ctxt, 'old', old_specs)
         new_type_ref = volume_types.create(self.ctxt, 'new', new_specs)
-        diff = {}
-        host = {}
+        diff = volume_types.volume_types_diff(self.ctxt, old_type_ref['id'],
+                                              new_type_ref['id'])[0]
         ret = self.driver.retype(
-            self.ctxt, TEST_VOLUME[0], new_type_ref, diff, host)
-        self.assertFalse(ret)
+            self.ctxt, TEST_VOLUME[0], new_type, diff, host)
+        self.assertEqual(3, request.call_count)
+        self.assertTrue(ret)
+
+    @mock.patch.object(requests.Session, "request")
+    def test_migrate_volume(self, request):
+        request.return_value = FakeResponse(200, GET_LDEV_RESULT)
+        host = {
+            'capabilities': {
+                'location_info': {
+                    'storage_id': CONFIG_MAP['serial'],
+                    'pool_id': 30,
+                },
+            },
+        }
+        ret = self.driver.migrate_volume(self.ctxt, TEST_VOLUME[0], host)
+        self.assertEqual(2, request.call_count)
+        actual = (True, None)
+        self.assertTupleEqual(actual, ret)
+
+    @mock.patch.object(requests.Session, "request")
+    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
+    def test_migrate_volume_diff_pool(
+            self, get_volume_type_extra_specs, request):
+        get_volume_type_extra_specs.return_value = {}
+        request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(200, GET_SNAPSHOTS_RESULT),
+                               FakeResponse(200, NOTFOUND_RESULT),
+                               FakeResponse(200, NOTFOUND_RESULT),
+                               FakeResponse(200, NOTFOUND_RESULT),
+                               FakeResponse(200, NOTFOUND_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
+        host = {
+            'capabilities': {
+                'location_info': {
+                    'storage_id': CONFIG_MAP['serial'],
+                    'pool_id': 40,
+                },
+            },
+        }
+        ret = self.driver.migrate_volume(self.ctxt, TEST_VOLUME[0], host)
+        self.assertEqual(1, get_volume_type_extra_specs.call_count)
+        self.assertEqual(15, request.call_count)
+        actual = (True, {'provider_location': '1'})
+        self.assertTupleEqual(actual, ret)
 
     def test_backup_use_temp_snapshot(self):
         self.assertTrue(self.driver.backup_use_temp_snapshot())
@@ -1073,32 +1390,46 @@ class HBSDRESTFCDriverTest(test.TestCase):
         self.assertTupleEqual(actual, ret)
 
     @mock.patch.object(requests.Session, "request")
-    def test_create_group_from_src_volume(self, request):
+    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
+    def test_create_group_from_src_volume(
+            self, get_volume_type_extra_specs, request):
+        get_volume_type_extra_specs.return_value = {}
         request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(200, GET_SNAPSHOTS_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
+        self.driver.common._stats = {}
+        self.driver.common._stats['pools'] = [
+            {'location_info': {'pool_id': 30}}]
         ret = self.driver.create_group_from_src(
             self.ctxt, TEST_GROUP[1], [TEST_VOLUME[1]],
             source_group=TEST_GROUP[0], source_vols=[TEST_VOLUME[0]]
         )
+        self.assertEqual(1, get_volume_type_extra_specs.call_count)
         self.assertEqual(5, request.call_count)
         actual = (
             None, [{'id': TEST_VOLUME[1]['id'], 'provider_location': '1'}])
         self.assertTupleEqual(actual, ret)
 
     @mock.patch.object(requests.Session, "request")
-    def test_create_group_from_src_snapshot(self, request):
+    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
+    def test_create_group_from_src_snapshot(
+            self, get_volume_type_extra_specs, request):
+        get_volume_type_extra_specs.return_value = {}
         request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(200, GET_SNAPSHOTS_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
+        self.driver.common._stats = {}
+        self.driver.common._stats['pools'] = [
+            {'location_info': {'pool_id': 30}}]
         ret = self.driver.create_group_from_src(
             self.ctxt, TEST_GROUP[0], [TEST_VOLUME[0]],
             group_snapshot=TEST_GROUP_SNAP[0], snapshots=[TEST_SNAPSHOT[0]]
         )
+        self.assertEqual(1, get_volume_type_extra_specs.call_count)
         self.assertEqual(5, request.call_count)
         actual = (
             None, [{'id': TEST_VOLUME[0]['id'], 'provider_location': '1'}])
@@ -1129,17 +1460,24 @@ class HBSDRESTFCDriverTest(test.TestCase):
 
     @mock.patch.object(requests.Session, "request")
     @mock.patch.object(sqlalchemy_api, 'volume_get', side_effect=_volume_get)
+    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
     @mock.patch.object(volume_utils, 'is_group_a_cg_snapshot_type')
     def test_create_group_snapshot_non_cg(
-            self, is_group_a_cg_snapshot_type, volume_get, request):
+            self, is_group_a_cg_snapshot_type, get_volume_type_extra_specs,
+            volume_get, request):
         is_group_a_cg_snapshot_type.return_value = False
+        get_volume_type_extra_specs.return_value = {}
         request.side_effect = [FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(200, GET_SNAPSHOTS_RESULT)]
+        self.driver.common._stats = {}
+        self.driver.common._stats['pools'] = [
+            {'location_info': {'pool_id': 30}}]
         ret = self.driver.create_group_snapshot(
             self.ctxt, TEST_GROUP_SNAP[0], [TEST_SNAPSHOT[0]]
         )
+        self.assertEqual(1, get_volume_type_extra_specs.call_count)
         self.assertEqual(4, request.call_count)
         actual = (
             {'status': 'available'},
@@ -1151,18 +1489,25 @@ class HBSDRESTFCDriverTest(test.TestCase):
 
     @mock.patch.object(requests.Session, "request")
     @mock.patch.object(sqlalchemy_api, 'volume_get', side_effect=_volume_get)
+    @mock.patch.object(volume_types, 'get_volume_type_extra_specs')
     @mock.patch.object(volume_utils, 'is_group_a_cg_snapshot_type')
     def test_create_group_snapshot_cg(
-            self, is_group_a_cg_snapshot_type, volume_get, request):
+            self, is_group_a_cg_snapshot_type, get_volume_type_extra_specs,
+            volume_get, request):
         is_group_a_cg_snapshot_type.return_value = True
+        get_volume_type_extra_specs.return_value = {}
         request.side_effect = [FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(200, GET_SNAPSHOTS_RESULT_PAIR),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT),
                                FakeResponse(200, GET_SNAPSHOTS_RESULT)]
+        self.driver.common._stats = {}
+        self.driver.common._stats['pools'] = [
+            {'location_info': {'pool_id': 30}}]
         ret = self.driver.create_group_snapshot(
             self.ctxt, TEST_GROUP_SNAP[0], [TEST_SNAPSHOT[0]]
         )
+        self.assertEqual(1, get_volume_type_extra_specs.call_count)
         self.assertEqual(5, request.call_count)
         actual = (
             None,
@@ -1183,10 +1528,14 @@ class HBSDRESTFCDriverTest(test.TestCase):
                                FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
+                               FakeResponse(200, GET_LDEV_RESULT),
                                FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)]
         ret = self.driver.delete_group_snapshot(
             self.ctxt, TEST_GROUP_SNAP[0], [TEST_SNAPSHOT[0]])
-        self.assertEqual(10, request.call_count)
+        self.assertEqual(14, request.call_count)
         actual = (
             {'status': TEST_GROUP_SNAP[0]['status']},
             [{'id': TEST_SNAPSHOT[0]['id'], 'status': 'deleted'}]
@@ -1199,6 +1548,41 @@ class HBSDRESTFCDriverTest(test.TestCase):
         ret = self.driver.get_driver_options()
         actual = (hbsd_common.COMMON_VOLUME_OPTS +
                   hbsd_common.COMMON_PORT_OPTS +
+                  hbsd_common.COMMON_PAIR_OPTS +
+                  hbsd_common.COMMON_NAME_OPTS +
                   hbsd_rest.REST_VOLUME_OPTS +
-                  hbsd_rest_fc.FC_VOLUME_OPTS)
+                  hbsd_rest.REST_PAIR_OPTS +
+                  hbsd_rest_fc.FC_VOLUME_OPTS +
+                  hbsd_replication._REP_OPTS +
+                  hbsd_replication.COMMON_MIRROR_OPTS +
+                  hbsd_replication.ISCSI_MIRROR_OPTS +
+                  hbsd_replication.REST_MIRROR_OPTS +
+                  hbsd_replication.REST_MIRROR_API_OPTS +
+                  hbsd_replication.REST_MIRROR_SSL_OPTS)
         self.assertEqual(actual, ret)
+
+    @mock.patch.object(requests.Session, "request")
+    def test_is_modifiable_dr_value_new_dr_mode_disabled(self, request):
+        request.side_effect = [
+            FakeResponse(200, GET_LDEV_RESULT_PAIR_STATUS_TEST),
+            FakeResponse(200, GET_LDEV_RESULT_PAIR_STATUS_TEST),
+            FakeResponse(202, COMPLETED_SUCCEEDED_RESULT)
+        ]
+        host = {
+            'capabilities': {
+                'location_info': {
+                    'pool_id': 30,
+                },
+            },
+        }
+        new_type = {'extra_specs': {'hbsd:capacity_saving': 'disable'}}
+        old_specs = {'hbsd:capacity_saving': 'deduplication_compression'}
+        new_specs = {'hbsd:capacity_saving': 'disable'}
+        old_type_ref = volume_types.create(self.ctxt, 'old', old_specs)
+        new_type_ref = volume_types.create(self.ctxt, 'new', new_specs)
+        diff = volume_types.volume_types_diff(self.ctxt, old_type_ref['id'],
+                                              new_type_ref['id'])[0]
+        ret = self.driver.retype(
+            self.ctxt, TEST_VOLUME[0], new_type, diff, host)
+        self.assertEqual(3, request.call_count)
+        self.assertTrue(ret)
