@@ -337,7 +337,6 @@ class NetAppCmodeNfsDriver(nfs_base.NetAppNfsDriver,
         # Used for service state report
         data['replication_enabled'] = self.replication_enabled
 
-        self._spawn_clean_cache_job()
         self._stats = data
 
     def _get_pool_stats(self, filter_function=None, goodness_function=None):
@@ -639,6 +638,8 @@ class NetAppCmodeNfsDriver(nfs_base.NetAppNfsDriver,
                 copied = True
             elif (cache_copy and
                   self.configuration.netapp_copyoffload_tool_path):
+                volume['provider_location'] = volume_utils.extract_host(
+                    volume['host'], level='pool')
                 LOG.debug("Trying copy from cache using copy offload.")
                 self._copy_from_remote_cache(volume, image_id, cache_copy)
                 copied = True
