@@ -34,6 +34,7 @@ import requests
 import six
 from six.moves import http_client
 
+from cinder.common import constants
 from cinder import exception
 from cinder.i18n import _
 from cinder import objects
@@ -85,9 +86,9 @@ class DPLCommand(object):
                   'Accept': 'application/cdmi-container',
                   'x-cdmi-specification-version': '1.0.2'}
         # base64 encode the username and password
-        auth = base64.encodestring('%s:%s'
-                                   % (self.username,
-                                      self.password)).replace('\n', '')
+        auth = base64.encodebytes('%s:%s'
+                                  % (self.username,
+                                     self.password)).replace('\n', '')
         header['Authorization'] = 'Basic %s' % auth
 
         if not params:
@@ -612,14 +613,6 @@ class DPLVolume(object):
                              [http_client.OK, http_client.ACCEPTED,
                               http_client.CREATED])
 
-    def get_vg_list(self, vgtype=None):
-        method = 'GET'
-        if vgtype:
-            url = '/%s/?volume_group_type=%s' % (DPL_OBJ_VOLUMEGROUP, vgtype)
-        else:
-            url = '/%s/' % (DPL_OBJ_VOLUMEGROUP)
-        return self._execute(method, url, None, [http_client.OK])
-
     def get_vg(self, groupID):
         method = 'GET'
         url = '/%s/%s/' % (DPL_OBJ_VOLUMEGROUP, groupID)
@@ -1463,7 +1456,7 @@ class DPLCOMMONDriver(driver.CloneableImageVD,
             if ret == 0:
                 data['vendor_name'] = output['metadata']['vendor']
                 data['driver_version'] = output['metadata']['version']
-                data['storage_protocol'] = 'iSCSI'
+                data['storage_protocol'] = constants.ISCSI
                 data['location_info'] = location_info
                 data['consistencygroup_support'] = True
                 data['consistent_group_snapshot_enabled'] = True
