@@ -10,10 +10,12 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+from __future__ import annotations
+
 import binascii
 import traceback
 import typing
-from typing import Any, Dict, List, Optional, Tuple  # noqa: H301
+from typing import Any, Optional  # noqa: H301
 
 from castellan import key_manager
 import os_brick.initiator.connectors
@@ -107,6 +109,7 @@ class OnFailureRescheduleTask(flow_utils.CinderTask):
             exception.VolumeNotFound,
             exception.SnapshotNotFound,
             exception.VolumeTypeNotFound,
+            exception.ImageConversionNotAllowed,
             exception.ImageUnacceptable,
             exception.ImageTooBig,
             exception.InvalidSignatureImage,
@@ -143,7 +146,7 @@ class OnFailureRescheduleTask(flow_utils.CinderTask):
                           volume.id)
 
     def _reschedule(self, context, cause, request_spec, filter_properties,
-                    volume):
+                    volume) -> None:
         """Actions that happen during the rescheduling attempt occur here."""
 
         create_volume = self.scheduler_rpcapi.create_volume
@@ -167,8 +170,8 @@ class OnFailureRescheduleTask(flow_utils.CinderTask):
             # Stringify to avoid circular ref problem in json serialization
             retry_info['exc'] = traceback.format_exception(*cause.exc_info)
 
-        return create_volume(context, volume, request_spec=request_spec,
-                             filter_properties=filter_properties)
+        create_volume(context, volume, request_spec=request_spec,
+                      filter_properties=filter_properties)
 
     def _post_reschedule(self, volume):
         """Actions that happen after the rescheduling attempt occur here."""
@@ -394,9 +397,14 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
         self.db = db
         self.driver = driver
         self.image_volume_cache = image_volume_cache
-        self.message = message_api.API()
-        self.backup_api = backup_api.API()
-        self.backup_rpcapi = backup_rpcapi.BackupAPI()
+        self._message = None
+
+    @property
+    def message(self):
+        if self._message is None:
+            self._message = message_api.API()
+
+        return self._message
 
     def _handle_bootable_volume_glance_meta(self, context, volume,
                                             **kwargs):
@@ -519,7 +527,14 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
 
         LOG.debug('rekey volume %s', volume.name)
 
-        properties = volume_utils.brick_get_connector_properties(False, False)
+        # Rekeying writes very little data (KB), so doing a single pathed
+        # connection is more efficient, but it is less robust, because the
+        # driver could be returning a single path, and it happens to be down
+        # then the attach would fail where the multipathed connection would
+        # have succeeded.
+        properties = volume_utils.brick_get_connector_properties(
+            self.driver.configuration.use_multipath_for_image_xfer,
+            self.driver.configuration.enforce_multipath_for_image_xfer)
         LOG.debug("properties: %s", properties)
         attach_info = None
         model_update = {}
@@ -673,7 +688,7 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
                                                    volume_metadata)
 
     @staticmethod
-    def _extract_cinder_ids(urls: List[str]) -> List[str]:
+    def _extract_cinder_ids(urls: list[str]) -> list[str]:
         """Process a list of location URIs from glance
 
         :param urls: list of glance location URIs
@@ -704,7 +719,7 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
                             context: cinder_context.RequestContext,
                             volume: objects.Volume,
                             image_location,
-                            image_meta: Dict[str, Any]) -> Tuple[None, bool]:
+                            image_meta: dict[str, Any]) -> tuple[None, bool]:
         """Create a volume efficiently from an existing image.
 
         Returns a dict of volume properties eg. provider_location,
@@ -765,7 +780,7 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
                                     context: cinder_context.RequestContext,
                                     volume: objects.Volume,
                                     image_location,
-                                    image_meta: Dict[str, Any],
+                                    image_meta: dict[str, Any],
                                     image_service) -> dict:
         # TODO(harlowja): what needs to be rolled back in the clone if this
         # volume create fails?? Likely this should be a subflow or broken
@@ -801,7 +816,7 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
             internal_context: cinder_context.RequestContext,
             volume: objects.Volume,
             image_id: str,
-            image_meta: Dict[str, Any]) -> Tuple[None, bool]:
+            image_meta: dict[str, Any]) -> tuple[None, bool]:
         """Attempt to create the volume using the image cache.
 
         Best case this will simply clone the existing volume in the cache.
@@ -850,8 +865,8 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
                                    volume: objects.Volume,
                                    image_location: str,
                                    image_id: str,
-                                   image_meta: Dict[str, Any],
-                                   image_service) -> Tuple[Optional[dict],
+                                   image_meta: dict[str, Any],
+                                   image_service) -> tuple[Optional[dict],
                                                            bool]:
         assert self.image_volume_cache is not None
         internal_context = cinder_context.get_internal_tenant_context()
@@ -892,7 +907,7 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
             volume: objects.Volume,
             image_location,
             image_id: str,
-            image_meta: Dict[str, Any],
+            image_meta: dict[str, Any],
             image_service,
             update_cache: bool = False) -> Optional[dict]:
         # NOTE(e0ne): check for free space in image_conversion_dir before
@@ -1011,6 +1026,14 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
                             detail=
                             message_field.Detail.SIGNATURE_VERIFICATION_FAILED,
                             exception=err)
+                except exception.ImageConversionNotAllowed:
+                    with excutils.save_and_reraise_exception():
+                        self.message.create(
+                            context,
+                            message_field.Action.COPY_IMAGE_TO_VOLUME,
+                            resource_uuid=volume.id,
+                            detail=
+                            message_field.Detail.IMAGE_FORMAT_UNACCEPTABLE)
 
             if should_create_cache_entry:
                 # Update the newly created volume db entry before we clone it
@@ -1042,7 +1065,7 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
                            volume: objects.Volume,
                            image_location,
                            image_id: str,
-                           image_meta: Dict[str, Any],
+                           image_meta: dict[str, Any],
                            image_service,
                            **kwargs: Any) -> Optional[dict]:
         LOG.debug("Cloning %(volume_id)s from image %(image_id)s "
@@ -1117,7 +1140,7 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
                             context: cinder_context.RequestContext,
                             volume: objects.Volume,
                             backup_id: str,
-                            **kwargs) -> Tuple[Dict, bool]:
+                            **kwargs) -> tuple[dict, bool]:
         LOG.info("Creating volume %(volume_id)s from backup %(backup_id)s.",
                  {'volume_id': volume.id,
                   'backup_id': backup_id})
@@ -1138,7 +1161,8 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
             volume.update(model_update)
             volume.save()
 
-            backup_host = self.backup_api.get_available_backup_service_host(
+            backupapi = backup_api.API()
+            backup_host = backupapi.get_available_backup_service_host(
                 backup.host, backup.availability_zone)
             updates = {'status': fields.BackupStatus.RESTORING,
                        'restore_volume_id': volume.id,
@@ -1146,8 +1170,9 @@ class CreateVolumeFromSpecTask(flow_utils.CinderTask):
             backup.update(updates)
             backup.save()
 
-            self.backup_rpcapi.restore_backup(context, backup.host, backup,
-                                              volume.id)
+            backuprpcapi = backup_rpcapi.BackupAPI()
+            backuprpcapi.restore_backup(context, backup.host, backup,
+                                        volume.id)
             need_update_volume = False
 
         LOG.info("Created volume %(volume_id)s from backup %(backup_id)s "
